@@ -21,7 +21,7 @@ MA 02110-1301 USA
 /*
   nepim - network pipemeter
 
-  $Id: main.c,v 1.72 2008/03/27 20:57:19 evertonm Exp $
+  $Id: main.c,v 1.73 2008/05/20 13:17:29 evertonm Exp $
  */
 
 #include <stdlib.h>
@@ -167,6 +167,9 @@ static void usage(FILE *out, const char *prog_name)
           "server-only options:\n"
           "  -f [lib-name]  enable TCP wrapper filtering\n"
           "  -j addr-list   join multicast groups (defaults to empty)\n"
+          "                 to join an SSM (source,group) pair, use: -j source+group\n"
+	  "                 examples: -j fe80::1+ff01::1111\n"
+	  "                           -j 10.10.10.10+232.1.1.1,2000\n"
 	  "  -U path-list   bind UNIX stream sockets to specified addresses\n"
 #if 0
 	  "  -G path-list   bind UNIX datagram sockets to specified addresses\n"
