@@ -18,7 +18,7 @@ Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110-1301 USA
 *-GNU-GPL-END-*/
 
-/* $Id: server.c,v 1.82 2008/03/27 20:57:19 evertonm Exp $ */
+/* $Id: server.c,v 1.83 2008/05/20 11:20:49 evertonm Exp $ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -1031,7 +1031,10 @@ static int spawn_tcp_listener(const char *hostname, const char *portname)
                                       nepim_global.win_send,
 				      nodelay,
 				      nepim_global.tos,
-				      nepim_global.server_tcp_mss);
+				      nepim_global.server_tcp_mss,
+				      -1, /* iface_index */
+				      0,  /* ssm_src_addr */
+				      -1  /* ssm_src_addrlen */);
     if (sd < 0) {
       fprintf(stderr, 
               "%s %s: TCP listener socket failed for %s,%d: %d\n",
@@ -1091,7 +1094,10 @@ static int spawn_unix_listener(const char *path)
 				    nepim_global.win_send,
 				    nodelay,
 				    nepim_global.tos,
-				    nepim_global.server_tcp_mss);
+				    nepim_global.server_tcp_mss,
+				    -1, /* iface_index */
+				    0,  /* ssm_src_addr */
+				    -1  /* ssm_src_addrlen */);
   if (sd < 0) {
     fprintf(stderr, 
 	    "%s %s: UNIX stream listener socket failed for path=%s\n",
@@ -1111,8 +1117,10 @@ static int spawn_unix_listener(const char *path)
 }
 
 
-extern int nepim_udp_listener(const char *hostname,
-                              const char *portname, int join);
+extern int nepim_udp_listener(const char *ssm_source,
+			      const char *hostname,
+                              const char *portname,
+			      int join);
 
 static void parse_listeners(const char *list, int *udp, int *tcp, int join)
 {
@@ -1134,11 +1142,34 @@ static void parse_listeners(const char *list, int *udp, int *tcp, int join)
     if (addr_split_port(buf, sizeof(buf), &portname))
       portname = (char *) nepim_global.portname;
 
-    *tcp += spawn_tcp_listener(buf, portname);
-    *udp += nepim_udp_listener(buf, portname, join);
-  }
-}
+    /*
+      buf:      host
+      portname: port
+     */
 
+    *tcp += spawn_tcp_listener(buf, portname);
+
+    if (join) {
+      char *group;
+      char *ssm_source;
+
+      if (addr_split_sourcegroup(buf, sizeof(buf), &group)) {
+	ssm_source = 0;
+	group = buf;
+      }
+      else {
+	ssm_source = buf;
+      }
+      *udp += nepim_udp_listener(ssm_source, group, portname, 1 /* join */);
+
+      continue; /* for */
+    }
+
+    *udp += nepim_udp_listener(0, buf, portname, 0 /* join */);
+
+  } /* for */
+}
+ 
 static int parse_unix_listeners(const char *list)
 {
   int listeners = 0;
@@ -1280,9 +1311,11 @@ void nepim_server_run()
       tcp_listeners += spawn_tcp_listener(INET6_ANY, nepim_global.portname);
     tcp_listeners += spawn_tcp_listener(INET_ANY, nepim_global.portname);
     if (!nepim_global.no_inet6)
-      udp_listeners += nepim_udp_listener(INET6_ANY, nepim_global.portname,
+      udp_listeners += nepim_udp_listener(0 /* ssm_source */,
+					  INET6_ANY, nepim_global.portname,
                                           0 /* no join */);
-    udp_listeners += nepim_udp_listener(INET_ANY, nepim_global.portname,
+    udp_listeners += nepim_udp_listener(0 /* ssm_source */,
+					INET_ANY, nepim_global.portname,
                                         0 /* no join */);
   }
 
