@@ -18,7 +18,7 @@ Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110-1301 USA
 *-GNU-GPL-END-*/
 
-/* $Id: sock.c,v 1.44 2008/03/27 20:57:19 evertonm Exp $ */
+/* $Id: sock.c,v 1.46 2008/05/20 11:44:34 evertonm Exp $ */
 
 #include <assert.h>
 #include <netdb.h>
@@ -55,6 +55,9 @@ MA 02110-1301 USA
 #ifndef IPV6_ADD_MEMBERSHIP
 #define IPV6_ADD_MEMBERSHIP 20
 #endif
+#ifndef MCAST_JOIN_SOURCE_GROUP
+#define MCAST_JOIN_SOURCE_GROUP 46
+#endif
 
 #ifndef HAVE_IP_MREQN
 struct ip_mreqn {
@@ -71,6 +74,14 @@ struct ipv6_mreq {
 };
 #endif
 
+#ifndef HAVE_GROUP_SOURCE_REQ
+struct group_source_req {
+  uint32_t                gsr_interface; 
+  struct sockaddr_storage gsr_group;
+  struct sockaddr_storage gsr_source;
+};
+#endif /* HAVE_GROUP_SOURCE_REQ */
+
 #define NEPIM_SOCK_ERR_NONE           (0)
 #define NEPIM_SOCK_ERR_UNSPEC         (-1)
 #define NEPIM_SOCK_ERR_SOCKET         (-2)
@@ -86,11 +97,12 @@ struct ipv6_mreq {
 #define NEPIM_SOCK_ERR_TTL            (-12)
 #define NEPIM_SOCK_ERR_MCAST_TTL      (-13)
 #define NEPIM_SOCK_ERR_MCAST_JOIN     (-14)
-#define NEPIM_SOCK_ERR_WIN_RECV       (-15)
-#define NEPIM_SOCK_ERR_WIN_SEND       (-16)
-#define NEPIM_SOCK_ERR_TOS            (-17)
-#define NEPIM_SOCK_ERR_SOCK_KA        (-18)
-#define NEPIM_SOCK_ERR_MSS            (-19)
+#define NEPIM_SOCK_ERR_MCAST_JOIN_SOURCEGROUP (-15)
+#define NEPIM_SOCK_ERR_WIN_RECV               (-16)
+#define NEPIM_SOCK_ERR_WIN_SEND               (-17)
+#define NEPIM_SOCK_ERR_TOS                    (-18)
+#define NEPIM_SOCK_ERR_SOCK_KA                (-19)
+#define NEPIM_SOCK_ERR_MSS                    (-20)
 
 int nepim_sock_get_port(const struct sockaddr *addr)
 {
@@ -392,7 +404,7 @@ static int create_socket(FILE *err, int domain, int type, int protocol,
 }
 
 static int socket_mcast_join(int sd, int family,
-			     const struct sockaddr *addr, int addr_len)
+			     const struct sockaddr *addr)
 {
   union {
     struct sockaddr_in inet;
@@ -434,6 +446,46 @@ static int socket_mcast_join(int sd, int family,
   return NEPIM_SOCK_ERR_MCAST_JOIN;
 }
 
+static int socket_mcast_join_sourcegroup(int sd, int family, int iface_index,
+					 const struct sockaddr *src,
+					 const struct sockaddr *grp)
+{
+  struct group_source_req group_source_req;
+
+  group_source_req.gsr_interface = iface_index;
+
+  switch (family) {
+  case PF_INET:
+    {
+      assert(sizeof(group_source_req.gsr_source) >= sizeof(struct sockaddr_in));
+      assert(sizeof(group_source_req.gsr_group) >= sizeof(struct sockaddr_in));
+
+      memcpy(&group_source_req.gsr_source, src, sizeof(struct sockaddr_in));
+      memcpy(&group_source_req.gsr_group, grp, sizeof(struct sockaddr_in));
+
+      return setsockopt(sd, SOL_IP, MCAST_JOIN_SOURCE_GROUP,
+			&group_source_req, sizeof(group_source_req));
+    }
+  case PF_INET6:
+    {
+      assert(sizeof(group_source_req.gsr_source) >= sizeof(struct sockaddr_in6));
+      assert(sizeof(group_source_req.gsr_group) >= sizeof(struct sockaddr_in6));
+
+      memcpy(&group_source_req.gsr_source, src, sizeof(struct sockaddr_in6));
+      memcpy(&group_source_req.gsr_group, grp, sizeof(struct sockaddr_in6));
+
+      return setsockopt(sd, SOL_IPV6, MCAST_JOIN_SOURCE_GROUP,
+			&group_source_req, sizeof(group_source_req));
+    }
+  default:
+    assert(0);
+  }
+
+  assert(0);
+
+  return NEPIM_SOCK_ERR_MCAST_JOIN_SOURCEGROUP;
+}
+
 int nepim_create_socket(FILE *err,
 			const struct sockaddr *addr, int addr_len,
 			int family,
@@ -446,7 +498,8 @@ int nepim_create_socket(FILE *err,
 			int win_send,
 			int nodelay,
 			int tos,
-			int mss)
+			int mss,
+			int iface, const struct sockaddr *ssm_src_addr, int ssm_src_addrlen)
 {
   int sd;
   int result;
@@ -456,11 +509,48 @@ int nepim_create_socket(FILE *err,
   if (sd < 0)
     return sd;
 
+#if 0
+      {
+	  char buf1[500];
+	  char buf2[500];
+	  nepim_sock_dump_addr(buf1, sizeof(buf1), ssm_src_addr, ssm_src_addrlen);
+	  nepim_sock_dump_addr(buf2, sizeof(buf2), addr, addr_len);
+	  fprintf(stderr, "DEBUG %s %s: mcast_join=%d mcast_join_sourcegroup(%d,%s,%s,%d)\n",
+		  __FILE__, __PRETTY_FUNCTION__,
+		  mcast_join, sd, buf1, buf2, nepim_sock_get_port(addr));
+      }
+#endif
+
   if (mcast_join) {
     assert(type == SOCK_DGRAM);
     assert(protocol == IPPROTO_UDP);
+
+    /* SSM join (S,G) ? */
+    if (ssm_src_addr) {
+
+      if (socket_mcast_join_sourcegroup(sd, family, iface, ssm_src_addr, addr)) {
+	
+	if (err) {
+	  char buf1[500];
+	  char buf2[500];
+	  int e = errno;
+	  
+	  nepim_sock_dump_addr(buf1, sizeof(buf1), ssm_src_addr, ssm_src_addrlen);
+	  nepim_sock_dump_addr(buf2, sizeof(buf2), addr, addr_len);
+	  
+	  fprintf(err, "%s: %s: mcast_join_sourcegroup(%d,%s,%s,%d): errno=%d: %s\n",
+		  __FILE__, __PRETTY_FUNCTION__,
+		  sd, buf1, buf2, nepim_sock_get_port(addr),
+		  e, strerror(e));
+	}
+	
+	close(sd);
+	return NEPIM_SOCK_ERR_MCAST_JOIN_SOURCEGROUP;
+      }
+    } /* ssm_src_addr */
     
-    if (socket_mcast_join(sd, family, addr, addr_len)) {
+      /* Regular mcast join (*,G) */
+    else if (socket_mcast_join(sd, family, addr)) {
 
       if (err) {
 	char buf[500];
@@ -528,14 +618,16 @@ int nepim_create_listener_socket(FILE *err,
 				 int win_send,
 				 int nodelay,
 				 int tos,
-				 int mss)
+				 int mss,
+				 int iface, const struct sockaddr *ssm_src_addr, int ssm_src_addrlen)
 {
   int sd;
 
   sd = nepim_create_socket(err, addr, addr_len, family,
 			   type, protocol, pmtu_mode, 
 			   ttl, 0, win_recv, win_send,
-			   nodelay, tos, mss);
+			   nodelay, tos, mss,
+			   iface, ssm_src_addr, ssm_src_addrlen);
   if (sd < 0)
     return sd;
 
@@ -745,7 +837,10 @@ int nepim_connect_client_socket(FILE *err,
   sd = nepim_create_socket(err, bind, bind_len,
 			   family, type, protocol, pmtu_mode, ttl,
 			   0 /* mcast_join: false */,
-			   win_recv, win_send, nodelay, tos, mss);
+			   win_recv, win_send, nodelay, tos, mss,
+			   -1, /* iface */
+			   0,  /* ssm_src_addr */
+			   -1  /* ssm_src_addrlen */);
   if (sd < 0)
     return sd;
 
