@@ -50,7 +50,8 @@ int nepim_create_socket(FILE *err,
 			int win_send,
 			int nodelay,
 			int tos,
-			int mss);
+			int mss,
+			int iface, const struct sockaddr *ssm_src_addr, int ssm_src_addrlen);
 int nepim_create_listener_socket(FILE *err,
 				 struct sockaddr *addr,
 				 int addr_len,
@@ -64,7 +65,8 @@ int nepim_create_listener_socket(FILE *err,
 				 int win_send,
 				 int nodelay,
 				 int tos,
-				 int mss);
+				 int mss,
+				 int iface, const struct sockaddr *ssm_src_addr, int ssm_src_addrlen);
 int nepim_connect_client_socket(FILE *err,
 				const struct sockaddr *bind, int bind_len,
 				struct sockaddr *addr, int addr_len,
