@@ -18,7 +18,7 @@ Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110-1301 USA
 *-GNU-GPL-END-*/
 
-/* $Id: udp_server.c,v 1.108 2008/03/27 20:57:20 evertonm Exp $ */
+/* $Id: udp_server.c,v 1.109 2008/05/20 11:20:49 evertonm Exp $ */
 
 
 #include <stdio.h>
@@ -1088,8 +1088,12 @@ static void *on_udp_segment(oop_source *src, int sd,
   return OOP_CONTINUE;
 }
 
-int nepim_udp_listener(const char *hostname, const char *portname, 
-		       int mcast_join)
+static int udp_listener(const char *ssm_source,
+			const struct sockaddr *ssm_source_addr,
+			int ssm_source_addrlen,
+			const char *hostname,
+			const char *portname, 
+			int mcast_join)
 {
   struct addrinfo hints;
   struct addrinfo *ai_res;
@@ -1097,6 +1101,13 @@ int nepim_udp_listener(const char *hostname, const char *portname,
   int result;
   int udp_listeners = 0;
 
+#if 0
+  fprintf(stderr,
+	  "DEBUG %s %s: mcast_join=%d ssm_source=%s host/group=%s port=%s\n",
+	  __FILE__, __PRETTY_FUNCTION__,
+	  mcast_join, ssm_source, hostname, portname);
+#endif
+
   memset(&hints, 0, sizeof(hints));
 
   hints.ai_socktype = SOCK_DGRAM;
@@ -1109,13 +1120,14 @@ int nepim_udp_listener(const char *hostname, const char *portname,
 
   result = getaddrinfo(hostname, portname, &hints, &ai_res);
   if (result) {
-    fprintf(stderr, "%s: getaddrinfo(%s,%s): %s\n",
-	    __PRETTY_FUNCTION__,
+    fprintf(stderr, "%s %s: getaddrinfo(%s,%s): %s\n",
+	    __FILE__, __PRETTY_FUNCTION__,
             hostname, portname, gai_strerror(result));
     return 0;
   }
 
   for (ai = ai_res; ai; ai = ai->ai_next) {
+    char ssm_source_addrstr[500];
     char addr_str[500];
     int sd;
 
@@ -1125,9 +1137,20 @@ int nepim_udp_listener(const char *hostname, const char *portname,
     if (nepim_global.no_inet4 && (ai->ai_family == PF_INET))
       continue;
 
-    nepim_sock_dump_addr(addr_str, sizeof addr_str,
+    if (ssm_source)
+      nepim_sock_dump_addr(ssm_source_addrstr, sizeof(ssm_source_addrstr),
+			   ssm_source_addr, ssm_source_addrlen);
+
+    nepim_sock_dump_addr(addr_str, sizeof(addr_str),
 			 ai->ai_addr, ai->ai_addrlen);
 
+#if 0
+  fprintf(stderr,
+	  "DEBUG %s %s: ssm_addr=%s host/group_addr=%s\n",
+	  __FILE__, __PRETTY_FUNCTION__,
+	  ssm_source_addrstr, addr_str);
+#endif
+
     sd = nepim_create_socket(nepim_global.verbose_stderr,
 			     ai->ai_addr, ai->ai_addrlen,
 			     ai->ai_family, ai->ai_socktype, 
@@ -1139,12 +1162,22 @@ int nepim_udp_listener(const char *hostname, const char *portname,
 			     nepim_global.win_send,
 			     0 /* tcp nodelay */,
 			     nepim_global.tos,
-			     nepim_global.server_tcp_mss);
+			     nepim_global.server_tcp_mss,
+			     0 /* "any" interface index */,
+			     ssm_source_addr,
+			     ssm_source_addrlen);
     if (sd < 0) {
-      fprintf(stderr, 
-              "%s %s: UDP listener socket failed for %s,%d: %d\n",
-              __FILE__, __PRETTY_FUNCTION__,
-	      addr_str, nepim_sock_get_port(ai->ai_addr), sd);
+      if (ssm_source)
+	fprintf(stderr, 
+		"%s %s: UDP listener socket failed for %s+%s,%d: %d\n",
+		__FILE__, __PRETTY_FUNCTION__,
+		ssm_source_addrstr,
+		addr_str, nepim_sock_get_port(ai->ai_addr), sd);
+      else
+	fprintf(stderr, 
+		"%s %s: UDP listener socket failed for %s,%d: %d\n",
+		__FILE__, __PRETTY_FUNCTION__,
+		addr_str, nepim_sock_get_port(ai->ai_addr), sd);
       break;
     }
 
@@ -1156,9 +1189,15 @@ int nepim_udp_listener(const char *hostname, const char *portname,
 
     ++udp_listeners;
 
-    fprintf(stderr, 
-	    "%d: UDP socket listening on %s,%d\n",
-	    sd, addr_str, nepim_sock_get_port(ai->ai_addr));
+    if (ssm_source)
+	fprintf(stderr, 
+		"%d: UDP socket listening on %s+%s,%d\n",
+		sd, ssm_source_addrstr,
+		addr_str, nepim_sock_get_port(ai->ai_addr));
+      else
+	fprintf(stderr, 
+		"%d: UDP socket listening on %s,%d\n",
+		sd, addr_str, nepim_sock_get_port(ai->ai_addr));
 
     nepim_sock_show_opt(nepim_global.verbose_stderr, stderr, sd);
   }
@@ -1168,6 +1207,66 @@ int nepim_udp_listener(const char *hostname, const char *portname,
   return udp_listeners;
 }
 
+int nepim_udp_listener(const char *ssm_source,
+		       const char *hostname,
+		       const char *portname, 
+		       int mcast_join)
+{
+  struct addrinfo hints;
+  struct addrinfo *ai_res;
+  struct addrinfo *ai;
+  int result;
+  int udp_listeners = 0;
+
+  if (!ssm_source)
+    return udp_listener(0  /* ssm_source */,
+			0  /* ssm_source_addr */,
+			-1 /* ssm_source_addrlen */,
+			hostname,
+			portname,
+			mcast_join);
+
+  memset(&hints, 0, sizeof(hints));
+
+  hints.ai_socktype = SOCK_DGRAM;
+  hints.ai_protocol = IPPROTO_UDP;
+  hints.ai_flags = AI_PASSIVE | AI_CANONNAME;
+  hints.ai_family = PF_UNSPEC;
+  hints.ai_addrlen = 0;
+  hints.ai_addr = 0;
+  hints.ai_canonname = 0;
+
+  result = getaddrinfo(ssm_source, 0, &hints, &ai_res);
+  if (result) {
+    fprintf(stderr, "%s: getaddrinfo(%s): %s\n",
+	    __PRETTY_FUNCTION__,
+            ssm_source, gai_strerror(result));
+    return 0;
+  }
+
+  /* Scan ssm_source addresses */
+
+  for (ai = ai_res; ai; ai = ai->ai_next) {
+
+    if (nepim_global.no_inet6 && (ai->ai_family == PF_INET6))
+      continue;
+
+    if (nepim_global.no_inet4 && (ai->ai_family == PF_INET))
+      continue;
+
+    udp_listeners += udp_listener(ssm_source,
+				  ai->ai_addr,
+				  ai->ai_addrlen,
+				  hostname,
+				  portname,
+				  mcast_join);
+  }
+
+  freeaddrinfo(ai_res);
+
+  return udp_listeners;
+}
+
 int nepim_unix_dgram_listener(const char *path)
 {
   struct sockaddr_un un_sock_addr;
@@ -1200,7 +1299,10 @@ int nepim_unix_dgram_listener(const char *path)
 			   nepim_global.win_send,
 			   0 /* tcp nodelay */,
 			   nepim_global.tos,
-			   nepim_global.server_tcp_mss);
+			   nepim_global.server_tcp_mss,
+			   -1 /* ssm iface index */,
+			   0  /* ssm_source_addr */,
+			   -1 /* ssm_source_addrlen */);
   if (sd < 0) {
     fprintf(stderr, 
 	    "%s %s: UNIX datagram listener socket failed for path=%s\n",
